@@ -1,53 +1,63 @@
-import os
-import json
-import time
-import requests  # type: ignore
+"""
+Local implementation of CodeBox.
+This is useful for testing and development.c
+In case you don't put an api_key,
+this is the default CodeBox.
+"""
+
 import asyncio
-import aiohttp
+import json
+import os
 import subprocess
+import time
+from asyncio.subprocess import Process
+from typing import List, Optional, Union
 from uuid import uuid4
-from typing import Optional, Union
+
+import aiohttp
+import requests
 from typing_extensions import Self
+from websockets.client import WebSocketClientProtocol
+from websockets.client import connect as ws_connect
 from websockets.exceptions import ConnectionClosedError
-from websockets.client import WebSocketClientProtocol, connect as ws_connect
-from websockets.sync.client import connect as ws_connect_sync, ClientConnection
+from websockets.sync.client import ClientConnection
+from websockets.sync.client import connect as ws_connect_sync
+
 from codeboxapi.box import BaseBox
-from codeboxapi.schema import (
-    CodeBoxStatus, 
-    CodeBoxOutput, 
-    CodeBoxFile
-)
+from codeboxapi.schema import CodeBoxFile, CodeBoxOutput, CodeBoxStatus
+
 from ..config import settings
 
 
 class LocalBox(BaseBox):
     """
     LocalBox is a CodeBox implementation that runs code locally.
-    This is useful for testing and development.c
-    In case you don't put an api_key, 
-    this is the default CodeBox.
+    This is useful for testing and development.
     """
+
     _instance: Optional[Self] = None
-    
+
     def __new__(cls, *args, **kwargs):
         if not cls._instance:
             cls._instance = super().__new__(cls, *args, **kwargs)
         else:
-            print(
-                "INFO: Using a LocalBox which is not isolated.\n"
-                "      This is only for testing and development.\n"
-                "      Make sure to put an API-Key in production.\n"
-            )
+            if settings.SHOW_INFO:
+                print(
+                    "INFO: Using a LocalBox which is not fully isolated\n"
+                    "      and not scalable across multiple users.\n"
+                    "      Make sure to use a CODEBOX_API_KEY in production.\n"
+                    "      Set envar SHOW_INFO=False to not see this again.\n"
+                )
         return cls._instance
-        
-    def __init__(self, port: int = 8888) -> None:
+
+    def __init__(self) -> None:
         super().__init__()
-        self.port = port
-        self.kernel: Optional[dict] = None
+        self.port: int = 8888
+        self.kernel_id: Optional[dict] = None
         self.ws: Union[WebSocketClientProtocol, ClientConnection, None] = None
-        self.subprocess: Union[asyncio.subprocess.Process, subprocess.Popen, None] = None
+        self.jupyter: Union[Process, subprocess.Popen, None] = None
         self.session: Optional[aiohttp.ClientSession] = None
-    
+
     def start(self) -> CodeBoxStatus:
         os.makedirs(".codebox", exist_ok=True)
         self._check_port()
@@ -56,50 +66,58 @@ class LocalBox(BaseBox):
             out = None
         else:
             out = subprocess.PIPE
-        self.subprocess = subprocess.Popen(
-            [
-                "jupyter",
-                "kernelgateway",
-                "--KernelGatewayApp.ip='0.0.0.0'",
-                f"--KernelGatewayApp.port={self.port}",
-            ],
-            stdout=out,
-            stderr=out,
-            cwd=".codebox",
-        )
+        try:
+            self.jupyter = subprocess.Popen(
+                [
+                    "jupyter",
+                    "kernelgateway",
+                    "--KernelGatewayApp.ip='0.0.0.0'",
+                    f"--KernelGatewayApp.port={self.port}",
+                ],
+                stdout=out,
+                stderr=out,
+                cwd=".codebox",
+            )
+        except FileNotFoundError:
+            raise ModuleNotFoundError(
+                "Jupyter Kernel Gateway not found, please install it with:\n"
+                "`pip install jupyter_kernel_gateway`\n"
+                "to use the LocalBox."
+            )
         while True:
             try:
-                response = requests.get(self.kernel_url)
-                if response.status_code == 200: break
+                response = requests.get(self.kernel_url, timeout=90)
+                if response.status_code == 200:
+                    break
             except requests.exceptions.ConnectionError:
                 pass
             if settings.VERBOSE:
                 print("Waiting for kernel to start...")
             time.sleep(1)
-        
+
         response = requests.post(
-            f"{self.kernel_url}/kernels", headers={"Content-Type": "application/json"}
+            f"{self.kernel_url}/kernels",
+            headers={"Content-Type": "application/json"},
+            timeout=90,
         )
-        self.kernel = response.json()
-        if self.kernel is None:
+        self.kernel_id = response.json()["id"]
+        if self.kernel_id is None:
             raise Exception("Could not start kernel")
-        
-        self.ws = ws_connect_sync(
-            f"{self.ws_url}/kernels/{self.kernel['id']}/channels"
-        )
-        
+
+        self.ws = ws_connect_sync(f"{self.ws_url}/kernels/{self.kernel_id}/channels")
+
         return CodeBoxStatus(status="started")
-    
+
     def _check_port(self) -> None:
         try:
-            response = requests.get(f"http://localhost:{self.port}")
+            response = requests.get(f"http://localhost:{self.port}", timeout=90)
         except requests.exceptions.ConnectionError:
             pass
         else:
             if response.status_code == 200:
                 self.port += 1
                 self._check_port()
-    
+
     async def astart(self) -> CodeBoxStatus:
         os.makedirs(".codebox", exist_ok=True)
         self.session = aiohttp.ClientSession()
@@ -109,19 +127,20 @@ class LocalBox(BaseBox):
             out = None
         else:
             out = asyncio.subprocess.PIPE
-        self.subprocess = await asyncio.create_subprocess_exec(
+        self.jupyter = await asyncio.create_subprocess_exec(
             "jupyter",
             "kernelgateway",
             "--KernelGatewayApp.ip='0.0.0.0'",
             f"--KernelGatewayApp.port={self.port}",
             stdout=out,
             stderr=out,
-            cwd=".codebox"
+            cwd=".codebox",
         )
         while True:
             try:
                 response = await self.session.get(self.kernel_url)
-                if response.status == 200: break
+                if response.status == 200:
+                    break
             except aiohttp.ClientConnectorError:
                 pass
             except aiohttp.ServerDisconnectedError:
@@ -133,15 +152,13 @@ class LocalBox(BaseBox):
         response = await self.session.post(
             f"{self.kernel_url}/kernels", headers={"Content-Type": "application/json"}
         )
-        self.kernel = await response.json()
-        if self.kernel is None:
+        self.kernel_id = (await response.json())["id"]
+        if self.kernel_id is None:
             raise Exception("Could not start kernel")
-        self.ws = await ws_connect(
-            f"{self.ws_url}/kernels/{self.kernel['id']}/channels"
-        )
-        
+        self.ws = await ws_connect(f"{self.ws_url}/kernels/{self.kernel_id}/channels")
+
         return CodeBoxStatus(status="started")
-    
+
     async def _acheck_port(self) -> None:
         try:
             if self.session is None:
@@ -155,45 +172,51 @@ class LocalBox(BaseBox):
             if response.status == 200:
                 self.port += 1
                 await self._acheck_port()
-            
+
     def status(self) -> CodeBoxStatus:
         return CodeBoxStatus(
-            status = "running" 
-                if self.kernel 
-                    and requests.get(self.kernel_url).status_code == 200
-                else "stopped"
+            status="running"
+            if self.kernel_id
+            and requests.get(self.kernel_url, timeout=90).status_code == 200
+            else "stopped"
         )
-    
+
     async def astatus(self) -> CodeBoxStatus:
         return CodeBoxStatus(
-            status = "running" 
-                if self.kernel 
-                    and self.session
-                    and (await self.session.get(self.kernel_url)).status == 200
-                else "stopped"
+            status="running"
+            if self.kernel_id
+            and self.session
+            and (await self.session.get(self.kernel_url)).status == 200
+            else "stopped"
         )
-    
-    def run(self, code: str | None = None, file_path: Optional[os.PathLike] = None, retry=3) -> CodeBoxOutput:
+
+    def run(
+        self,
+        code: Optional[str] = None,
+        file_path: Optional[os.PathLike] = None,
+        retry=3,
+    ) -> CodeBoxOutput:
         if not code and not file_path:
             raise ValueError("Code or file_path must be specified!")
-        
+
         if code and file_path:
             raise ValueError("Can only specify code or the file to read_from!")
-        
+
         if file_path:
-            with open(file_path, 'r') as f:
+            with open(file_path, "r", encoding="utf-8") as f:
                 code = f.read()
 
         # run code in jupyter kernel
-        if retry <= 0: 
+        if retry <= 0:
             raise RuntimeError("Could not connect to kernel")
-        if not self.ws: 
+        if not self.ws:
             self.start()
-            if not self.ws: raise RuntimeError("Could not connect to kernel")
-        
+            if not self.ws:
+                raise RuntimeError("Could not connect to kernel")
+
         if settings.VERBOSE:
             print("Running code:\n", code)
-        
+
         # send code to kernel
         self.ws.send(
             json.dumps(
@@ -225,8 +248,8 @@ class LocalBox(BaseBox):
                 received_msg = json.loads(self.ws.recv())
             except ConnectionClosedError:
                 self.start()
-                return self.run(code, file_path, retry-1)
-            
+                return self.run(code, file_path, retry - 1)
+
             if (
                 received_msg["header"]["msg_type"] == "stream"
                 and received_msg["parent_header"]["msg_id"] == msg_id
@@ -237,7 +260,7 @@ class LocalBox(BaseBox):
                 result += msg + "\n"
                 if settings.VERBOSE:
                     print("Output:\n", result)
-            
+
             elif (
                 received_msg["header"]["msg_type"] == "execute_result"
                 and received_msg["parent_header"]["msg_id"] == msg_id
@@ -245,66 +268,75 @@ class LocalBox(BaseBox):
                 result += received_msg["content"]["data"]["text/plain"].strip() + "\n"
                 if settings.VERBOSE:
                     print("Output:\n", result)
-            
+
             elif received_msg["header"]["msg_type"] == "display_data":
                 if "image/png" in received_msg["content"]["data"]:
                     return CodeBoxOutput(
-                        type="image/png", 
-                        content=received_msg["content"]["data"]["image/png"]
+                        type="image/png",
+                        content=received_msg["content"]["data"]["image/png"],
                     )
-                elif "text/plain" in received_msg["content"]["data"]:
+                if "text/plain" in received_msg["content"]["data"]:
                     return CodeBoxOutput(
-                        type="text", 
-                        content=received_msg["content"]["data"]["text/plain"]
+                        type="text",
+                        content=received_msg["content"]["data"]["text/plain"],
                     )
+                return CodeBoxOutput(
+                    type="error",
+                    content="Could not parse output",
+                )
             elif (
                 received_msg["header"]["msg_type"] == "status"
                 and received_msg["parent_header"]["msg_id"] == msg_id
                 and received_msg["content"]["execution_state"] == "idle"
-            ):  
-                if len(result) > 500: 
+            ):
+                if len(result) > 500:
                     result = "[...]\n" + result[-500:]
                 return CodeBoxOutput(
-                    type = "text", 
-                    content = result or "code run successfully (no output)"
+                    type="text", content=result or "code run successfully (no output)"
                 )
 
             elif (
                 received_msg["header"]["msg_type"] == "error"
                 and received_msg["parent_header"]["msg_id"] == msg_id
             ):
-                error = f"{received_msg['content']['ename']}: {received_msg['content']['evalue']}"
+                error = (
+                    f"{received_msg['content']['ename']}: "
+                    f"{received_msg['content']['evalue']}"
+                )
                 if settings.VERBOSE:
                     print("Error:\n", error)
-                return CodeBoxOutput(
-                    type="error", 
-                    content=error
-                )
-    
-    async def arun(self, code: str | None = None, file_path: Optional[os.PathLike] = None, retry=3) -> CodeBoxOutput:
-        if not code and not file_path:
-            raise ValueError("Code or file_path must be specified!")
-        
-        if code and file_path:
-            raise ValueError("Can only specify code or the file to read_from!")
-        
+                return CodeBoxOutput(type="error", content=error)
+
+            return CodeBoxOutput(
+                type="error",
+                content="Unknown message",
+            )
+
+    async def arun(
+        self,
+        code: str,
+        file_path: Optional[os.PathLike] = None,
+        retry=3,
+    ) -> CodeBoxOutput:
         if file_path:
-            with open(file_path, 'r') as f:
-                code = await f.read()
+            raise NotImplementedError(
+                "Reading from file is not supported in async mode"
+            )
 
         # run code in jupyter kernel
-        if retry <= 0: 
+        if retry <= 0:
             raise RuntimeError("Could not connect to kernel")
-        if not self.ws: 
+        if not self.ws:
             await self.astart()
-            if not self.ws: raise RuntimeError("Could not connect to kernel")
-        
+            if not self.ws:
+                raise RuntimeError("Could not connect to kernel")
+
         if settings.VERBOSE:
             print("Running code:\n", code)
-        
+
         if not isinstance(self.ws, WebSocketClientProtocol):
             raise RuntimeError("Mixing asyncio and sync code is not supported")
-        
+
         await self.ws.send(
             json.dumps(
                 {
@@ -333,8 +365,8 @@ class LocalBox(BaseBox):
                 received_msg = json.loads(await self.ws.recv())
             except ConnectionClosedError:
                 await self.astart()
-                return await self.arun(code, file_path, retry-1)
-            
+                return await self.arun(code, file_path, retry - 1)
+
             if (
                 received_msg["header"]["msg_type"] == "stream"
                 and received_msg["parent_header"]["msg_id"] == msg_id
@@ -345,7 +377,7 @@ class LocalBox(BaseBox):
                 result += msg + "\n"
                 if settings.VERBOSE:
                     print("Output:\n", result)
-            
+
             elif (
                 received_msg["header"]["msg_type"] == "execute_result"
                 and received_msg["parent_header"]["msg_id"] == msg_id
@@ -353,83 +385,93 @@ class LocalBox(BaseBox):
                 result += received_msg["content"]["data"]["text/plain"].strip() + "\n"
                 if settings.VERBOSE:
                     print("Output:\n", result)
-            
+
             elif received_msg["header"]["msg_type"] == "display_data":
                 if "image/png" in received_msg["content"]["data"]:
                     return CodeBoxOutput(
-                        type="image/png", 
-                        content=received_msg["content"]["data"]["image/png"]
+                        type="image/png",
+                        content=received_msg["content"]["data"]["image/png"],
                     )
-                elif "text/plain" in received_msg["content"]["data"]:
+                if "text/plain" in received_msg["content"]["data"]:
                     return CodeBoxOutput(
-                        type="text", 
-                        content=received_msg["content"]["data"]["text/plain"]
+                        type="text",
+                        content=received_msg["content"]["data"]["text/plain"],
                     )
             elif (
                 received_msg["header"]["msg_type"] == "status"
                 and received_msg["parent_header"]["msg_id"] == msg_id
                 and received_msg["content"]["execution_state"] == "idle"
-            ):  
-                if len(result) > 500: 
+            ):
+                if len(result) > 500:
                     result = "[...]\n" + result[-500:]
                 return CodeBoxOutput(
-                    type = "text", 
-                    content = result or "code run successfully (no output)"
+                    type="text", content=result or "code run successfully (no output)"
                 )
 
             elif (
                 received_msg["header"]["msg_type"] == "error"
                 and received_msg["parent_header"]["msg_id"] == msg_id
             ):
-                error = f"{received_msg['content']['ename']}: {received_msg['content']['evalue']}"
+                error = (
+                    f"{received_msg['content']['ename']}: "
+                    f"{received_msg['content']['evalue']}"
+                )
                 if settings.VERBOSE:
                     print("Error:\n", error)
-                return CodeBoxOutput(
-                    type="error", 
-                    content=error
-                )
-        
+                return CodeBoxOutput(type="error", content=error)
+
     def upload(self, file_name: str, content: bytes) -> CodeBoxStatus:
         os.makedirs(".codebox", exist_ok=True)
         with open(os.path.join(".codebox", file_name), "wb") as f:
             f.write(content)
-        
+
         return CodeBoxStatus(status=f"{file_name} uploaded successfully")
-    
+
     async def aupload(self, file_name: str, content: bytes) -> CodeBoxStatus:
         return await asyncio.to_thread(self.upload, file_name, content)
-    
+
     def download(self, file_name: str) -> CodeBoxFile:
         with open(os.path.join(".codebox", file_name), "rb") as f:
             content = f.read()
-        
-        return CodeBoxFile(
-            name=file_name,
-            content=content
-        )
-        
+
+        return CodeBoxFile(name=file_name, content=content)
+
     async def adownload(self, file_name: str) -> CodeBoxFile:
         return await asyncio.to_thread(self.download, file_name)
-        
+
     def install(self, package_name: str) -> CodeBoxStatus:
         self.run(f"!pip install -q {package_name}")
         # restart kernel if needed TODO
         return CodeBoxStatus(status=f"{package_name} installed successfully")
-    
+
     async def ainstall(self, package_name: str) -> CodeBoxStatus:
         await self.arun(f"!pip install -q {package_name}")
         # restart kernel if needed TODO
         return CodeBoxStatus(status=f"{package_name} installed successfully")
-    
-    def list_files(self) -> list[CodeBoxFile]:
+
+    def list_files(self) -> List[CodeBoxFile]:
         return [
             CodeBoxFile(name=file_name, content=None)
             for file_name in os.listdir(".codebox")
         ]
-        
-    async def alist_files(self) -> list[CodeBoxFile]:
+
+    async def alist_files(self) -> List[CodeBoxFile]:
         return await asyncio.to_thread(self.list_files)
-    
+
+    def restart(self) -> CodeBoxStatus:
+        if self.jupyter is not None:
+            self.stop()
+        else:
+            self.start()
+        return CodeBoxStatus(status="restarted")
+
+    async def arestart(self) -> CodeBoxStatus:
+        if self.jupyter is not None:
+            await self.astop()
+        else:
+            await self.astart()
+        return CodeBoxStatus(status="restarted")
+
     def stop(self) -> CodeBoxStatus:
         if self.ws is not None:
             try:
@@ -437,16 +479,16 @@ class LocalBox(BaseBox):
             except ConnectionClosedError:
                 pass
             self.ws = None
-        
-        if self.subprocess is not None:
-            self.subprocess.terminate()
-            self.subprocess.wait()
-            self.subprocess = None
+
+        if self.jupyter is not None:
+            self.jupyter.terminate()
+            self.jupyter.wait()
+            self.jupyter = None
             time.sleep(2)
-            
+
         return CodeBoxStatus(status="stopped")
-    
-    async def astop(self) -> CodeBoxStatus:            
+
+    async def astop(self) -> CodeBoxStatus:
         if self.ws is not None:
             try:
                 if not isinstance(self.ws, WebSocketClientProtocol):
@@ -455,22 +497,24 @@ class LocalBox(BaseBox):
             except ConnectionClosedError:
                 pass
             self.ws = None
-            
-        if self.subprocess is not None:
-            self.subprocess.terminate()
-            self.subprocess = None
+
+        if self.jupyter is not None:
+            self.jupyter.terminate()
+            self.jupyter = None
             await asyncio.sleep(2)
-            
+
         if self.session is not None:
             await self.session.close()
             self.session = None
-        
+
         return CodeBoxStatus(status="stopped")
 
     @property
     def kernel_url(self) -> str:
+        """Return the url of the kernel."""
         return f"http://localhost:{self.port}/api"
-    
+
     @property
     def ws_url(self) -> str:
+        """Return the url of the websocket."""
         return f"ws://localhost:{self.port}/api"
